@@ -6,10 +6,9 @@ import warnings
 import torch
 from torch.utils.data import DataLoader
 from torch.utils.data.distributed import DistributedSampler
-from tqdm import tqdm
 
 from tez import enums
-from tez.callbacks import CallbackRunner
+from tez.callbacks import CallbackRunner, Progress
 from tez.logger import logger
 from tez.utils import AverageMeter
 
@@ -37,6 +36,7 @@ class Tez:
         self.world_size = 1
         self._model_state = None
         self._train_state = None
+        self.metrics_meter = None
         self.metrics = {}
         self.metrics["train"] = {}
         self.metrics["valid"] = {}
@@ -109,13 +109,23 @@ class Tez:
         else:
             self.valid_collate_fn = None
 
-        if "callbacks" in kwargs:
-            self.callbacks = kwargs["callbacks"]
+        num_train_steps = int(len(self.train_dataset) / self.config.training_batch_size * self.config.epochs)
+        if self.valid_dataset:
+            num_valid_steps = int(len(self.valid_dataset) / self.config.validation_batch_size)
         else:
-            self.callbacks = []
+            num_valid_steps = None
+
+        _progress = Progress(num_train_steps=num_train_steps, num_valid_steps=num_valid_steps)
+
+        if "callbacks" in kwargs:
+            self.callbacks = [_progress] + kwargs["callbacks"]
+        else:
+            self.callbacks = [_progress]
 
         if self.config.num_jobs == -1:
             self.config.num_jobs = multiprocessing.cpu_count()
+            if self.config.num_jobs > 4:
+                self.config.num_jobs -= 2
 
         if self.world_size > 1 and self.train_sampler is None:
             self.train_sampler = DistributedSampler(
@@ -165,13 +175,18 @@ class Tez:
         self._configure_model()
         self.train_state = enums.TrainingState.TRAIN_START
 
+        if self.optimizer is None:
+            raise Exception("No optimizer found")
+
         if self.local_rank != -1:
             if torch.distributed.get_rank() == 0:
                 logger.info(f"\n{self.config}")
-                if self.optimizer is None:
-                    raise Exception("No optimizer found")
                 if self.scheduler is None:
                     logger.warning("No scheduler found. Continuing without scheduler")
+        else:
+            logger.info(f"\n{self.config}")
+            if self.scheduler is None:
+                logger.warning("No scheduler found. Continuing without scheduler")
 
     def _init_load_weights(self, config):
         self.config = config
@@ -217,6 +232,8 @@ class Tez:
         return self.metrics[v_1][v_2]
 
     def update_metrics(self, losses, monitor):
+        if self._model_state.value == "end":
+            return
         self.metrics[self._model_state.value].update(monitor)
         self.metrics[self._model_state.value]["loss"] = losses.avg
 
@@ -278,10 +295,11 @@ class Tez:
             output, loss, metrics = self.model(**data)
         return output, loss, metrics
 
-    def train_step(self, data):
+    def _zero_grad(self):
         if self.config.gradient_accumulation_steps == 1 and self.batch_index == 0:
             self.model.zero_grad()
-        _, loss, metrics = self.model_fn(data)
+
+    def _backward(self, loss, metrics):
         if self.num_gpu > 1:
             loss = loss.mean()
             for metric in metrics:
@@ -293,10 +311,14 @@ class Tez:
         else:
             loss.backward()
 
+    def _clip_grad_norm(self):
         if self.config.clip_grad_norm != -1:
             torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.config.clip_grad_norm)
 
-        if (self.batch_index + 1) % self.config.gradient_accumulation_steps == 0:
+    def _step(self):
+        is_bi_mod_acc_zero = (self.batch_index + 1) % self.config.gradient_accumulation_steps == 0
+        is_bi_end = self.batch_index + 1 == self.train_loader
+        if is_bi_mod_acc_zero or is_bi_end:
             if self.config.fp16:
                 self.scaler.step(self.optimizer)
                 self.scaler.update()
@@ -311,9 +333,14 @@ class Tez:
                         step_metric = self.name_to_metric(self.config.step_scheduler_metric)
                         self.scheduler.step(step_metric)
 
-            if self.batch_index > 0:
-                self.model.zero_grad()
+            self.model.zero_grad()
 
+    def train_step(self, data):
+        self._zero_grad()
+        _, loss, metrics = self.model_fn(data)
+        self._backward(loss, metrics)
+        self._clip_grad_norm()
+        self._step()
         return loss, metrics
 
     def predict_step(self, data):
@@ -324,88 +351,108 @@ class Tez:
                 metrics[metric] = metrics[metric].mean()
         return loss, metrics
 
-    def train(self, data_loader, _tqdm=None):
+    def _set_training_epoch_start(self, data_loader):
         if isinstance(data_loader, DataLoader) and isinstance(data_loader.sampler, DistributedSampler):
             data_loader.sampler.set_epoch(self.current_epoch)
 
-        self.train_state = enums.TrainingState.EPOCH_START
+        self.model_state = enums.ModelState.TRAIN
         self.train_state = enums.TrainingState.TRAIN_EPOCH_START
         self.model.train()
-        self.model_state = enums.ModelState.TRAIN
-        losses = AverageMeter()
         if self.config.gradient_accumulation_steps > 1:
             self.optimizer.zero_grad()
 
+    def _update_loss_metrics(self, losses, loss, metrics, data_loader):
+        if self.model_state == enums.ModelState.TRAIN:
+            losses.update(loss.item() * self.config.gradient_accumulation_steps, data_loader.batch_size)
+        else:
+            losses.update(loss.item(), data_loader.batch_size)
+
+        if self.batch_index == 0:
+            self.metrics_meter = {k: AverageMeter() for k in metrics}
+
+        monitor = {}
+        for m_m in self.metrics_meter:
+            self.metrics_meter[m_m].update(metrics[m_m].cpu().detach().numpy(), data_loader.batch_size)
+            monitor[m_m] = self.metrics_meter[m_m].avg
+        if self.model_state == enums.ModelState.TRAIN:
+            self.current_train_step += 1
+        else:
+            self.current_valid_step += 1
+        self.update_metrics(losses=losses, monitor=monitor)
+        return losses, monitor
+
+    def _set_training_epoch_end(self, losses, monitor):
+        self.update_metrics(losses=losses, monitor=monitor)
+        self.train_state = enums.TrainingState.TRAIN_EPOCH_END
+
+    def _set_training_state(self):
+        self.model_state = enums.ModelState.TRAIN
+        self.train_state = enums.TrainingState.TRAIN_EPOCH_START
+        self.model.train()
+
+    def train(self, data_loader):
+        self._set_training_epoch_start(data_loader)
+        losses = AverageMeter()
         for batch_index, data in enumerate(data_loader):
             self.batch_index = batch_index
             self.train_state = enums.TrainingState.TRAIN_STEP_START
             loss, metrics = self.train_step(data)
+            losses, monitor = self._update_loss_metrics(losses, loss, metrics, data_loader)
             self.train_state = enums.TrainingState.TRAIN_STEP_END
-            losses.update(loss.item() * self.config.gradient_accumulation_steps, data_loader.batch_size)
-            if batch_index == 0:
-                metrics_meter = {k: AverageMeter() for k in metrics}
-            monitor = {}
-            for m_m in metrics_meter:
-                metrics_meter[m_m].update(metrics[m_m].cpu().detach().numpy(), data_loader.batch_size)
-                monitor[m_m] = metrics_meter[m_m].avg
-            self.current_train_step += 1
-            if _tqdm is not None:
-                _tqdm.set_postfix(loss=losses.avg, stage="train", epoch=self.current_epoch, **monitor)
-                _tqdm.update(1)
-        self.update_metrics(losses=losses, monitor=monitor)
-        self.train_state = enums.TrainingState.TRAIN_EPOCH_END
-        return losses.avg
+            if self.valid_loader and self.config.val_strategy == "batch":
+                if self.current_train_step % self.config.val_steps == 0:
+                    self.validate(self.valid_loader)
+            if self._model_state.value == "end":
+                break
+        self._set_training_epoch_end(losses, monitor)
 
-    def validate(self, data_loader, _tqdm=None):
+    def _set_validation_epoch_start(self, data_loader):
         if isinstance(data_loader, DataLoader) and isinstance(data_loader.sampler, DistributedSampler):
             data_loader.sampler.set_epoch(self.current_epoch)
+
         self.train_state = enums.TrainingState.VALID_EPOCH_START
-        self.model.eval()
         self.model_state = enums.ModelState.VALID
+        self.model.eval()
+
+    def _set_validation_epoch_end(self, losses, monitor):
+        self.update_metrics(losses=losses, monitor=monitor)
+        self.train_state = enums.TrainingState.VALID_EPOCH_END
+
+    def validate(self, data_loader):
+        self._set_validation_epoch_start(data_loader)
         losses = AverageMeter()
 
         for batch_index, data in enumerate(data_loader):
+            self.batch_index = batch_index
             self.train_state = enums.TrainingState.VALID_STEP_START
             with torch.no_grad():
                 loss, metrics = self.predict_step(data)
+            losses, monitor = self._update_loss_metrics(losses, loss, metrics, data_loader)
             self.train_state = enums.TrainingState.VALID_STEP_END
-            losses.update(loss.item(), data_loader.batch_size)
-            if batch_index == 0:
-                metrics_meter = {k: AverageMeter() for k in metrics}
-            monitor = {}
-            for m_m in metrics_meter:
-                metrics_meter[m_m].update(metrics[m_m].cpu().detach().numpy(), data_loader.batch_size)
-                monitor[m_m] = metrics_meter[m_m].avg
-            if _tqdm is not None:
-                _tqdm.set_postfix(loss=losses.avg, stage="valid", epoch=self.current_epoch, **monitor)
-            self.current_valid_step += 1
-        self.update_metrics(losses=losses, monitor=monitor)
-        self.train_state = enums.TrainingState.VALID_EPOCH_END
-        return losses.avg
+        self._set_validation_epoch_end(losses, monitor)
+        if self.config.val_strategy == "batch" and self._model_state.value != "end":
+            self._set_training_state()
+
+    def _step_scheduler_after_epoch(self):
+        if self.scheduler is not None:
+            if self.config.step_scheduler_after == "epoch":
+                if self.config.step_scheduler_metric is None:
+                    self.scheduler.step()
+                else:
+                    step_metric = self.name_to_metric(self.config.step_scheduler_metric)
+                    self.scheduler.step(step_metric)
 
     def fit(self, train_dataset, valid_dataset=None, config: TezConfig = None, **kwargs):
         if config is None:
             config = TezConfig()
         self._init_trainer(train_dataset, valid_dataset, config, **kwargs)
-        num_train_steps = int(
-            len(self.train_dataset)
-            / self.config.training_batch_size
-            / self.config.gradient_accumulation_steps
-            * self.config.epochs
-        )
-        _tqdm = tqdm(total=num_train_steps)
-        for _ in range(self.config.epochs):
-            _ = self.train(self.train_loader, _tqdm)
-            if self.valid_loader:
-                _ = self.validate(self.valid_loader)
 
-            if self.scheduler is not None:
-                if self.config.step_scheduler_after == "epoch":
-                    if self.config.step_scheduler_metric is None:
-                        self.scheduler.step()
-                    else:
-                        step_metric = self.name_to_metric(self.config.step_scheduler_metric)
-                        self.scheduler.step(step_metric)
+        for _ in range(self.config.epochs):
+            self.train_state = enums.TrainingState.EPOCH_START
+            self.train(self.train_loader)
+            if self.valid_loader and self.config.val_strategy == "epoch":
+                self.validate(self.valid_loader)
+            self._step_scheduler_after_epoch()
             self.train_state = enums.TrainingState.EPOCH_END
             if self._model_state.value == "end":
                 time.sleep(2)
@@ -449,6 +496,8 @@ class Tez:
 
         if num_jobs == -1:
             num_jobs = multiprocessing.cpu_count()
+            if num_jobs > 4:
+                num_jobs -= 2
 
         if batch_size == 1:
             num_jobs = 0
@@ -465,14 +514,8 @@ class Tez:
         if self.model.training:
             self.model.eval()
 
-        tk0 = tqdm(data_loader, total=len(data_loader))
-
-        for data in tk0:
+        for data in data_loader:
             with torch.no_grad():
                 out, _, _ = self.model_fn(data)
                 out = self.process_output(out)
                 yield out
-
-            tk0.set_postfix(stage="test")
-
-        tk0.close()
